"""Model selection."""

import numpy as np

from .utils import shuffle_dataset


def split_dataset(
    X: np.ndarray, y: np.ndarray, train: float = 0.7, shuffle: bool = True
):
    """Split the dataset into train/test sets."""
    if train < 0 or train > 1:
        raise ValueError("train proportion must be between 0 and 1")

    if shuffle:
        X, y = shuffle_dataset(X, y)

    index = round(train * X.shape[0])

    return X[:index], y[:index], X[index:], y[index:]



"""Classifiers."""

import numpy as np

from abc import ABC, abstractmethod

from .metrics import accuracy_score
from .optimizers import GradientDescent
from .preprocessors import Standardization
from .utils import check_dataset_consistency


class AbstractClassifier(ABC):
    """Abstract class for classifiers algorithms."""

    def __init__(self, *args, optimizer=None, standardize: bool = True, **kwargs):
        self.standardize = Standardization() if standardize else None
        self.optimizer = optimizer if optimizer else GradientDescent(*args, **kwargs)
        self.optimizer.build(self.cost, self.gradient)
        self.parameters = None
        self.labels = None

    @abstractmethod
    def hypothesis(self, X, theta):
        """Hypothesis."""
        pass  # pragma: no cover

    @abstractmethod
    def cost(self, X, y, theta):
        """Cost function."""
        pass  # pragma: no cover

    @abstractmethod
    def gradient(self, X, y, theta):
        """Gradient."""
        pass  # pragma: no cover

    @property
    def history(self):
        return self.optimizer.history

    def fit(self, X, y):
        """Fit the model."""
        if not check_dataset_consistency(X, y):
            raise ValueError("the features set and target set must have as many rows")

        if self.standardize is not None:
            X = self.standardize(X)
        X = np.concatenate((np.ones((X.shape[0], 1)), X), axis=1)

        self.labels = np.unique(y)
        labels_number = np.size(self.labels)
        if labels_number < 2:
            raise ValueError("target must have at least two different classes")
        elif labels_number == 2:
            self.parameters = self.optimizer(X, y)
        else:
            self.parameters = self.optimizer(X, (y == self.labels).astype(int))

    def predict_probability(self, X):
        """Predict the probability of a target given features."""
        if self.standardize is not None:
            X = self.standardize(X)
        X = np.concatenate((np.ones((X.shape[0], 1)), X), axis=1)
        return self.hypothesis(X, self.parameters)

    def predict(self, X):
        """Predict a target given features."""
        probability = self.predict_probability(X)
        if np.size(probability, axis=1) > 1:
            return self.labels[np.argmax(probability, axis=1).reshape(-1, 1)]
        return self.labels[np.around(probability).astype("int")]

    def score(self, X, y):
        """Score of the model."""
        return accuracy_score(self.predict(X), y)


class LinearClassifier(AbstractClassifier):
    """Linear classifier (logistic regressor)."""

    def sigmoid(self, z):
        """Logistic function."""
        return 1 / (1 + np.exp(-z))

    def hypothesis(self, X, theta):
        """Logistic hypothesis."""
        return self.sigmoid(np.dot(X, theta))

    def cost(self, X, y, theta):
        """Cost function."""
        return (
            -y.T.dot(np.log(self.hypothesis(X, theta)))
            - (1 - y).T.dot(np.log(1 - self.hypothesis(X, theta)))
        ).flat[0]

    def gradient(self, X, y, theta):
        """Gradient."""
        return X.T.dot(self.hypothesis(X, theta) - y)


class RidgeClassifier(LinearClassifier):
    """Regularized linear classifier."""

    def __init__(self, *args, regularization: float = 1, **kwargs):
        super().__init__(*args, **kwargs)
        self.regularization = regularization

    def sigmoid(self, z):
        """Logistic function."""
        return 1 / (1 + np.exp(-z))

    def hypothesis(self, X, theta):
        """Logistic hypothesis."""
        return self.sigmoid(np.dot(X, theta))

    def cost(self, X, y, theta):
        """Regularized cost function."""
        return (
            -y.T.dot(np.log(self.hypothesis(X, theta)))
            - (1 - y).T.dot(np.log(1 - self.hypothesis(X, theta)))
        ).flat[0] + self.regularization * np.sum(np.square(theta[:, 1:]), axis=0)

    def gradient(self, X, y, theta):
        """Regularized gradient."""
        return (
            X.T.dot(self.hypothesis(X, theta) - y)
            + self.regularization * np.c_[np.zeros((theta.shape[0], 1)), theta[:, 1:]]
        )



"""Utils."""

import numpy as np


from typing import Tuple


# --- Dataset ---


def check_dataset_consistency(X: np.ndarray, y: np.ndarray) -> bool:
    """Check the length consistency between the features set and the target set."""
    return X.shape[0] == y.shape[0]


def shuffle_dataset(X: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Randomly shuffle the dataset."""
    indices = np.random.permutation(X.shape[0])
    X = X[indices[:, np.newaxis], np.arange(X.shape[1])]
    y = y[indices[:, np.newaxis], np.arange(y.shape[1])]
    return X, y


# --- Target ---


def target_reshape(y: np.ndarray) -> np.ndarray:
    """Reshape the target to be a 1-D array."""
    return y.reshape(-1)


def target_labels(*args) -> np.ndarray:
    """Get the target labels."""
    return np.unique(np.vstack(args))


def target_label_count(*args) -> int:
    """Get the target label count."""
    label_count = len(target_labels(*args))
    return label_count if label_count >= 2 else 2


def target_label_to_index(y: np.ndarray, labels=None) -> np.ndarray:
    """Transform label to index."""
    if labels is None:
        labels = target_labels(y)

    translation = {label: index for index, label in enumerate(labels)}
    return np.vectorize(translation.get)(y)



"""Metrics tools."""

import numpy as np

from typing import Optional

from .utils import check_dataset_consistency
from .utils import (
    target_reshape,
    target_labels,
    target_label_count,
    target_label_to_index,
)


def confusion_matrix(y_pred, y_true):
    """Confusion matrix."""
    if not check_dataset_consistency(y_pred, y_true):
        raise ValueError("input must have as many rows")

    labels = target_labels(y_pred, y_true)

    y_pred = target_label_to_index(target_reshape(y_pred), labels=labels)
    y_true = target_label_to_index(target_reshape(y_true), labels=labels)

    label_count = len(labels)

    cm = np.zeros((label_count, label_count), dtype=int)

    for i in range(y_true.size):
        cm[y_true[i]][y_pred[i]] += 1

    return cm


def accuracy_score(y_pred, y_true, normalize: bool = True):
    """Accuracy score.

    **normalize**
    If set to True, it returns the fraction of correct predictions.
    Else, it returns the sum of correct predictions.
    """
    if not check_dataset_consistency(y_pred, y_true):
        raise ValueError("input must have as many rows")

    cm = confusion_matrix(y_pred, y_true)

    accuracies = np.diag(cm).astype(float)

    if not normalize:
        return np.sum(accuracies)
    return np.sum(accuracies) / len(target_reshape(y_true))


def precision_score(y_pred, y_true, average: Optional[str] = None):
    """Precision score.

    In the case of multiclass labels, an average strategy can be set:

    **none**
    The score for each label is returned.

    **macro**
    The mean of each label score is returned.
    """
    if not check_dataset_consistency(y_pred, y_true):
        raise ValueError("input must have as many rows")

    cm = confusion_matrix(y_pred, y_true)

    numerator = np.diag(cm).astype(float)
    denominator = np.sum(cm, axis=0).astype(float)

    precisions = np.divide(
        numerator, denominator, out=np.zeros_like(numerator), where=denominator != 0
    )

    if target_label_count(y_true) <= 2:
        return np.sum(precisions)
    elif not average:
        return precisions
    elif average == "macro":
        return np.mean(precisions)
    else:
        raise ValueError(f"average {average} not supported")


def recall_score(y_pred, y_true, average: Optional[str] = None):
    """Recall score.

    In the case of multiclass labels, an average strategy can be set:

    **none**
    The score for each label is returned.

    **macro**
    The mean of each label score is returned.
    """
    if not check_dataset_consistency(y_pred, y_true):
        raise ValueError("input must have as many rows")

    cm = confusion_matrix(y_pred, y_true)

    numerator = np.diag(cm).astype(float)
    denominator = np.sum(cm, axis=1).astype(float)

    recalls = np.divide(
        numerator, denominator, out=np.zeros_like(numerator), where=denominator != 0
    )

    if target_label_count(y_true) <= 2:
        return np.sum(recalls)
    elif not average:
        return recalls
    elif average == "macro":
        return np.mean(recalls)
    else:
        raise ValueError(f"average {average} not supported")


def fbeta_score(y_pred, y_true, beta, average: Optional[str] = None):
    """F-beta score.

    In the case of multiclass labels, an average strategy can be set:

    **none**
    The score for each label is returned.

    **macro**
    The mean of each label score is returned.
    """
    if not check_dataset_consistency(y_pred, y_true):
        raise ValueError("input must have as many rows")

    precisions = precision_score(y_pred, y_true, average=None)
    recalls = recall_score(y_pred, y_true, average=None)

    numerator = (1 + beta ** 2) * precisions * recalls
    denominator = (beta ** 2) * precisions + recalls

    f_beta_scores = np.divide(
        numerator, denominator, out=np.zeros_like(numerator), where=denominator != 0
    )

    if target_label_count(y_true) <= 2:
        return np.sum(f_beta_scores)
    elif not average:
        return f_beta_scores
    elif average == "macro":
        return np.mean(f_beta_scores)
    else:
        raise ValueError(f"average {average} not supported")


def f1_score(y_pred, y_true, average: Optional[str] = None):
    """F1 score.

    In the case of multiclass labels, an average strategy can be set:

    **none**
    The score for each label is returned.

    **macro**
    The mean of each label score is returned.
    """
    return fbeta_score(y_pred, y_true, 1, average=average)


def r2_score(y_pred, y_true):
    """Coefficient of determination score."""
    if not check_dataset_consistency(y_pred, y_true):
        raise ValueError("input must have as many rows")

    ss_res = np.sum(np.square(y_true - y_pred), axis=0)
    ss_total = np.sum(np.square(y_true - np.mean(y_true, axis=0)), axis=0)

    if ss_total == 0:
        return 1

    return np.mean((1 - (ss_res / ss_total)))



"""Regressors."""

import numpy as np

from abc import ABC, abstractmethod

from .metrics import r2_score
from .optimizers import GradientDescent
from .preprocessors import Standardization
from .utils import check_dataset_consistency


class AbstractRegressor(ABC):
    """Abstract class for regressors algorithms."""

    def __init__(self, *args, optimizer=None, standardize: bool = True, **kwargs):
        self.standardize = Standardization() if standardize else None
        self.optimizer = optimizer if optimizer else GradientDescent(*args, **kwargs)
        self.optimizer.build(self.cost, self.gradient)

    @abstractmethod
    def hypothesis(self, X, theta):
        """Hypothesis."""
        pass  # pragma: no cover

    @abstractmethod
    def cost(self, X, y, theta):
        """Cost function."""
        pass  # pragma: no cover

    @abstractmethod
    def gradient(self, X, y, theta):
        """Gradient."""
        pass  # pragma: no cover

    @property
    def parameters(self):
        return self.optimizer.parameters

    @property
    def history(self):
        return self.optimizer.history

    def fit(self, X, y):
        """Fit the model."""
        if not check_dataset_consistency(X, y):
            raise ValueError("the features set and target set must have as many rows")

        if self.standardize is not None:
            X = self.standardize(X)
        X = np.concatenate((np.ones((X.shape[0], 1)), X), axis=1)
        self.optimizer(X, y)

    def predict(self, X):
        """Predict a target given features."""
        if self.standardize is not None:
            X = self.standardize(X)
        X = np.concatenate((np.ones((X.shape[0], 1)), X), axis=1)
        return self.hypothesis(X, self.parameters)

    def score(self, X, y):
        """Score of the model."""
        return r2_score(self.predict(X), y)


class LinearRegressor(AbstractRegressor):
    """Linear regressor."""

    def hypothesis(self, X, theta):
        """Linear hypothesis."""
        return np.dot(X, theta)

    def cost(self, X, y, theta):
        """Cost function."""
        return (1 / 2) * (self.hypothesis(X, theta) - y).T.dot(
            self.hypothesis(X, theta) - y
        ).flat[0]

    def gradient(self, X, y, theta):
        """Gradient."""
        return X.T.dot(self.hypothesis(X, theta) - y)

    def normal(self, X, y):
        """Use normal equation to compute the parameters."""
        X = np.concatenate((np.ones((X.shape[0], 1)), X), axis=1)
        self.optimizer.parameters = np.linalg.pinv(X.T.dot(X)).dot(X.T).dot(y)


class RidgeRegressor(AbstractRegressor):
    """Ridge regressor."""

    def __init__(self, *args, regularization: float = 1, **kwargs):
        super().__init__(*args, **kwargs)
        self.regularization = regularization

    def hypothesis(self, X, theta):
        """Linear hypothesis."""
        return np.dot(X, theta)

    def cost(self, X, y, theta):
        """Regularized cost function."""
        return (1 / 2) * (self.hypothesis(X, theta) - y).T.dot(
            self.hypothesis(X, theta) - y
        ).flat[0] + self.regularization * np.sum(np.square(theta[:, 1:]), axis=0)

    def gradient(self, X, y, theta):
        """Regularized gradient."""
        return (
            X.T.dot(self.hypothesis(X, theta) - y)
            + self.regularization * np.c_[np.zeros((theta.shape[0], 1)), theta[:, 1:]]
        )

    def normal(self, X, y):
        """Use normal equation regularized to compute the parameters."""
        X = np.concatenate((np.ones((X.shape[0], 1)), X), axis=1)
        L = np.identity(X.shape[0])
        L[0, 0] = 0
        self.optimizer.parameters = (
            np.linalg.pinv(X.T.dot(X) + self.regularization * L).dot(X.T).dot(y)
        )



"""Preprocessors."""

import numpy as np

from typing import Optional


class Normalization(object):
    """Rescale the data via a normalization.

    Produce:
    - Bring all values into the range [0, 1]
    """

    def __call__(self, X, axis: int = 0):
        min_x = np.amin(X, axis=axis)
        max_x = np.amax(X, axis=axis)
        return (X - min_x) / (max_x - min_x)


class Standardization(object):
    """Rescale the data via a standardization

    Produce:
    - mean(Xstandardized) = 0
    - std(Xstandardized) = 1
    """

    def __init__(self, mu: Optional[int] = None, sigma: Optional[int] = None):
        self.mu = mu
        self.sigma = sigma

    def __call__(self, X, axis: int = 0):
        if self.mu is None or self.sigma is None:
            self.mu = np.mean(X, axis=axis)
            self.sigma = np.std(X, axis=axis)

            if not np.any(self.sigma):
                self.sigma = np.ones_like(self.sigma)

        return np.divide(X - self.mu, self.sigma)



"""Optimizers algorithms."""

import numpy as np

from abc import ABC, abstractmethod
from typing import Optional

from .utils import check_dataset_consistency, shuffle_dataset


class AbstractOptimizer(ABC):
    """Abstract class for optimizers algorithms."""

    def __init__(self, iterations: int = 100, history: bool = False):
        self.iterations = iterations

        self.function = None
        self.gradient = None

        self.is_history_enabled = history
        self.history: Optional[list] = [] if history else None

    def build(self, function, gradient):
        self.function = function
        self.gradient = gradient

    @abstractmethod
    def __call__(self):
        pass  # pragma: no cover


class GradientDescent(AbstractOptimizer):
    """Batch gradient descent."""

    def __init__(self, *args, learning_rate: float = 0.01, **kwargs):
        super().__init__(*args, **kwargs)
        self.learning_rate = learning_rate

        self.parameters = None

    def __call__(self, X, y, *args, **kwargs):
        if not (self.function and self.gradient):
            raise ValueError("you must build the optimizer before calling it")

        if not check_dataset_consistency(X, y):
            raise ValueError("the features set and target set must have as many rows")

        m = y.shape[0]

        self.parameters = np.zeros((X.shape[1], 1))
        for _ in range(self.iterations):
            self.parameters = self.parameters - (
                self.learning_rate / m
            ) * self.gradient(X, y, self.parameters, *args, **kwargs)

            if self.is_history_enabled:
                self.history.append((1 / m) * self.function(X, y, self.parameters))

        return self.parameters


class SGD(AbstractOptimizer):
    """Stochastic gradient descent."""

    def __init__(
        self, *args, learning_rate: float = 0.01, iterations: int = 1, **kwargs
    ):
        kwargs.update({"iterations": iterations})
        super().__init__(*args, **kwargs)
        self.learning_rate = learning_rate

        self.parameters = None

    def __call__(self, X, y, *args, **kwargs):
        if not (self.function and self.gradient):
            raise ValueError("you must build the optimizer before calling it")

        if not check_dataset_consistency(X, y):
            raise ValueError("the features set and target set must have as many rows")

        X, y = shuffle_dataset(X, y)

        self.parameters = np.zeros((X.shape[1], 1))
        for _ in range(self.iterations):
            for X_i, y_i in zip(X, y):
                self.parameters = self.parameters - self.learning_rate * self.gradient(
                    X_i[None, :], y_i[None, :], self.parameters, *args, **kwargs
                )

                if self.is_history_enabled:
                    self.history.append(self.function(X, y, self.parameters))

        return self.parameters


class MBGD(AbstractOptimizer):
    """Mini-batch gradient descent."""

    def __init__(
        self,
        *args,
        learning_rate: float = 0.01,
        batch_size: int = 10,
        iterations: int = 10,
        **kwargs,
    ):
        kwargs.update({"iterations": iterations})
        super().__init__(*args, **kwargs)
        self.learning_rate = learning_rate
        self.batch_size = batch_size

        self.parameters = None

    def __call__(self, X, y, *args, **kwargs):
        if not (self.function and self.gradient):
            raise ValueError("you must build the optimizer before calling it")

        if not check_dataset_consistency(X, y):
            raise ValueError("the features set and target set must have as many rows")

        X, y = shuffle_dataset(X, y)
        m = y.shape[0]

        self.parameters = np.zeros((X.shape[1], 1))
        for _ in range(self.iterations):
            for i in range(0, m, self.batch_size):
                self.parameters = self.parameters - (
                    self.learning_rate / m
                ) * self.gradient(
                    X[i : i + self.batch_size],
                    y[i : i + self.batch_size],
                    self.parameters,
                    *args,
                    **kwargs,
                )

            if self.is_history_enabled:
                self.history.append(
                    (1 / self.batch_size) * self.function(X, y, self.parameters)
                )

        return self.parameters



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['alchina']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.0,<4.0',
 'numpy>=1.16,<2.0',
 'pandas>=0.24.2,<0.25.0',
 'scikit-learn>=0.20.3,<0.21.0']

setup_kwargs = {
    'name': 'alchina',
    'version': '0.1.1',
    'description': 'Machine Learning framework',
    'long_description': '# Alchina /al.ki.na/\n\n[![Build Status](https://travis-ci.org/matthieugouel/alchina.svg?branch=master)](https://travis-ci.org/matthieugouel/alchina)\n[![Coverage Status](https://coveralls.io/repos/github/matthieugouel/alchina/badge.svg?branch=master)](https://coveralls.io/github/matthieugouel/alchina?branch=master)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n[![license](https://img.shields.io/github/license/matthieugouel/gibica.svg)](https://github.com/matthieugouel/gibica/blob/master/LICENSE)\n\nAlchina is a Machine Learning framework.\n',
    'author': 'matthieu',
    'author_email': 'matthieu.gouel@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
